"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const errorTypes_1 = require("../errorTypes");
const pnpm_logger_1 = require("pnpm-logger");
const loadYamlFile = require("load-yaml-file");
const writeYamlFile = require("write-yaml-file");
const R = require("ramda");
const rimraf = require("rimraf-then");
const isCI = require("is-ci");
const getRegistryName_1 = require("../resolve/npm/getRegistryName");
const pnpmPkgJson_1 = require("../pnpmPkgJson");
const shrinkwrapLogger = pnpm_logger_1.default('shrinkwrap');
exports.SHRINKWRAP_FILENAME = 'shrinkwrap.yaml';
exports.PRIVATE_SHRINKWRAP_FILENAME = path.join('node_modules', '.shrinkwrap.yaml');
const SHRINKWRAP_VERSION = 3;
const CREATED_WITH = `${pnpmPkgJson_1.default.name}@${pnpmPkgJson_1.default.version}`;
class ShrinkwrapBreakingChangeError extends errorTypes_1.PnpmError {
    constructor(filename) {
        super('SHRINKWRAP_BREAKING_CHANGE', `Shrinkwrap file ${filename} not compatible with current pnpm`);
        this.filename = filename;
    }
}
function getDefaultShrinkwrap(registry) {
    return {
        version: SHRINKWRAP_VERSION,
        createdWith: CREATED_WITH,
        specifiers: {},
        dependencies: {},
        packages: {},
        registry,
    };
}
function readPrivate(pkgPath, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const shrinkwrapPath = path.join(pkgPath, exports.PRIVATE_SHRINKWRAP_FILENAME);
        let shrinkwrap;
        try {
            shrinkwrap = yield loadYamlFile(shrinkwrapPath);
        }
        catch (err) {
            if (err.code !== 'ENOENT') {
                throw err;
            }
            return getDefaultShrinkwrap(opts.registry);
        }
        if (shrinkwrap && shrinkwrap.version === SHRINKWRAP_VERSION) {
            return shrinkwrap;
        }
        if (opts.force || isCI) {
            shrinkwrapLogger.warn(`Ignoring not compatible shrinkwrap file at ${shrinkwrapPath}`);
            return getDefaultShrinkwrap(opts.registry);
        }
        throw new ShrinkwrapBreakingChangeError(shrinkwrapPath);
    });
}
exports.readPrivate = readPrivate;
function read(pkgPath, opts) {
    return __awaiter(this, void 0, void 0, function* () {
        const shrinkwrapPath = path.join(pkgPath, exports.SHRINKWRAP_FILENAME);
        let shrinkwrap;
        try {
            shrinkwrap = yield loadYamlFile(shrinkwrapPath);
        }
        catch (err) {
            if (err.code !== 'ENOENT') {
                throw err;
            }
            return getDefaultShrinkwrap(opts.registry);
        }
        if (shrinkwrap && shrinkwrap.version === SHRINKWRAP_VERSION) {
            return shrinkwrap;
        }
        if (opts.force || isCI) {
            shrinkwrapLogger.warn(`Ignoring not compatible shrinkwrap file at ${shrinkwrapPath}`);
            return getDefaultShrinkwrap(opts.registry);
        }
        throw new ShrinkwrapBreakingChangeError(shrinkwrapPath);
    });
}
exports.read = read;
function save(pkgPath, shrinkwrap) {
    const shrinkwrapPath = path.join(pkgPath, exports.SHRINKWRAP_FILENAME);
    const privateShrinkwrapPath = path.join(pkgPath, exports.PRIVATE_SHRINKWRAP_FILENAME);
    // empty shrinkwrap is not saved
    if (Object.keys(shrinkwrap.dependencies).length === 0) {
        return Promise.all([
            rimraf(shrinkwrapPath),
            rimraf(privateShrinkwrapPath),
        ]);
    }
    const formatOpts = {
        sortKeys: true,
        lineWidth: 1000,
        noCompatMode: true,
    };
    return Promise.all([
        writeYamlFile(shrinkwrapPath, shrinkwrap, formatOpts),
        writeYamlFile(privateShrinkwrapPath, shrinkwrap, formatOpts),
    ]);
}
exports.save = save;
function prune(shr, pkg) {
    const packages = {};
    const optionalDependencies = R.keys(pkg.optionalDependencies);
    const dependencies = R.difference(R.keys(pkg.dependencies), optionalDependencies);
    const devDependencies = R.difference(R.difference(R.keys(pkg.devDependencies), optionalDependencies), dependencies);
    copyDependencyTree(packages, shr, {
        registry: shr.registry,
        dependencies: devDependencies,
        dev: true,
    });
    copyDependencyTree(packages, shr, {
        registry: shr.registry,
        dependencies: optionalDependencies,
        optional: true,
    });
    copyDependencyTree(packages, shr, {
        registry: shr.registry,
        dependencies,
    });
    const allDeps = R.reduce(R.union, [], [optionalDependencies, devDependencies, dependencies]);
    const specifiers = {};
    const shrDependencies = {};
    R.keys(shr.specifiers).forEach(depName => {
        if (allDeps.indexOf(depName) === -1)
            return;
        specifiers[depName] = shr.specifiers[depName];
        shrDependencies[depName] = shr.dependencies[depName];
    });
    return {
        version: SHRINKWRAP_VERSION,
        createdWith: shr.createdWith || CREATED_WITH,
        specifiers,
        registry: shr.registry,
        dependencies: shrDependencies,
        packages,
    };
}
exports.prune = prune;
function copyDependencyTree(resolvedPackages, shr, opts) {
    let pkgIds = opts.dependencies
        .map((pkgName) => getPkgShortId(shr.dependencies[pkgName], pkgName));
    const checked = new Set();
    while (pkgIds.length) {
        let nextPkgIds = [];
        for (let pkgId of pkgIds) {
            if (checked.has(pkgId))
                continue;
            checked.add(pkgId);
            if (!shr.packages[pkgId]) {
                pnpm_logger_1.default.warn(`Cannot find resolution of ${pkgId} in shrinkwrap file`);
                continue;
            }
            const depShr = shr.packages[pkgId];
            resolvedPackages[pkgId] = depShr;
            if (opts.optional) {
                depShr.optional = true;
            }
            else {
                delete depShr.optional;
            }
            if (opts.dev) {
                depShr.dev = true;
            }
            else {
                delete depShr.dev;
            }
            const newDependencies = R.keys(depShr.dependencies)
                .map((pkgName) => getPkgShortId((depShr.dependencies && depShr.dependencies[pkgName]), pkgName))
                .filter((newPkgId) => !checked.has(newPkgId));
            nextPkgIds = R.union(nextPkgIds, newDependencies);
        }
        pkgIds = nextPkgIds;
    }
    return resolvedPackages;
}
function shortIdToFullId(shortId, registry) {
    if (shortId[0] === '/') {
        const registryName = getRegistryName_1.default(registry);
        return `${registryName}${shortId}`;
    }
    return shortId;
}
exports.shortIdToFullId = shortIdToFullId;
function getPkgShortId(reference, pkgName) {
    if (reference.indexOf('/') === -1) {
        return `/${pkgName}/${reference}`;
    }
    return reference;
}
exports.getPkgShortId = getPkgShortId;
function getPkgId(reference, pkgName, registry) {
    if (reference.indexOf('/') === -1) {
        const registryName = getRegistryName_1.default(registry);
        return `${registryName}/${pkgName}/${reference}`;
    }
    return reference;
}
exports.getPkgId = getPkgId;
function pkgIdToRef(pkgId, pkgName, resolution, standardRegistry) {
    if (resolution.type)
        return pkgId;
    const registryName = getRegistryName_1.default(standardRegistry);
    if (pkgId.startsWith(`${registryName}/`)) {
        const ref = pkgId.replace(`${registryName}/${pkgName}/`, '');
        if (ref.indexOf('/') === -1)
            return ref;
        return pkgId.replace(`${registryName}/`, '/');
    }
    return pkgId;
}
exports.pkgIdToRef = pkgIdToRef;
function pkgShortId(pkgId, standardRegistry) {
    const registryName = getRegistryName_1.default(standardRegistry);
    if (pkgId.startsWith(`${registryName}/`)) {
        return pkgId.substr(pkgId.indexOf('/'));
    }
    return pkgId;
}
exports.pkgShortId = pkgShortId;
//# sourceMappingURL=shrinkwrap.js.map